# -*- coding: utf-8 -*-

"""
Dropdown menu implementation.
"""

import typing as T

from .constants import SHOW_ITEMS_LIMIT, SCROLL_SPEED
from .item import T_ITEM
from .exc import NoItemToSelectError


class Dropdown:
    """
    Simulate an items dropdown menu. User can move or scroll selector up and down
    to select an item, and then perform :mod:`~zelfred.action`.

    For example, the ``[x] Beautiful is better than ugly.``,
    ``[ ] Explicit is better than implicit.`` and
    the ``[ ] Simple is better than complex.`` part in the following UI is the
    dropdown menu.

    .. code-block:: bash

        (Query):
        [x] Beautiful is better than ugly.
              subtitle 01
        [ ] Explicit is better than implicit.
              subtitle 02
        [ ] Simple is better than complex.
              subtitle 03

    :param items: All items in this dropdown menu. We only show ``SHOW_ITEMS_LIMIT``
        items in the UI at a time
    :param n_items: total number of items, it is a cache of ``len(items)``.
    :param selected_item_index: the selected item index, the value can be larger
        than ``SHOW_ITEMS_LIMIT`` if we have many items.
    :param cursor_position: the selected item cursor position in the dropdown UI,
        it is a value from 0 to ``SHOW_ITEMS_LIMIT - 1``.
    :param show_items_limit:  max number of items to show in the UI.
    """

    def __init__(self, items: T.List[T_ITEM]):
        self.items: T.List[T_ITEM] = items
        self.n_items: int = len(items)
        self.selected_item_index: int = 0
        self.cursor_position: int = 0
        self.show_items_limit = min(SHOW_ITEMS_LIMIT, self.n_items)

    def update(self, items: T.List[T_ITEM]):
        """
        Update the dropdown menu with new items.
        """
        self.items = items
        self.n_items = len(items)
        self.selected_item_index = 0
        self.cursor_position = 0
        self.show_items_limit = min(SHOW_ITEMS_LIMIT, self.n_items)

    @property
    def selected_item(self) -> T_ITEM:
        """
        Return the selected item object. If there is no item, raise
        :class:`~zelfred.exc.NoItemToSelectError`.
        """
        try:
            return self.items[self.selected_item_index]
        except IndexError:
            raise NoItemToSelectError

    def _press_down(self) -> T.Tuple[int, int]:
        # already the last item
        if self.selected_item_index == self.n_items - 1:
            select_delta = 0
        else:
            self.selected_item_index += 1
            select_delta = 1
        # already the last item in the UI
        if self.cursor_position == self.show_items_limit - 1:
            cursor_delta = 0
        else:
            self.cursor_position += 1
            cursor_delta = 1
        return select_delta, cursor_delta

    def press_down(self, n: int = 1) -> T.Tuple[int, int]:
        """
        Move selector down to pick the next item (if possible) in the dropdown menu.

        Example, before ``dropdown.press_down()``:

        .. code-block:: bash

            (Query):
            [x] Beautiful is better than ugly.
                  subtitle 01
            [ ] Explicit is better than implicit.
                  subtitle 02
            [ ] Simple is better than complex.
                  subtitle 03

        After:

        .. code-block:: bash

            (Query):
            [ ] Beautiful is better than ugly.
                  subtitle 01
            [x] Explicit is better than implicit.
                  subtitle 02
            [ ] Simple is better than complex.
                  subtitle 03

        :param n: move selector down ``n`` times.

        :return: tuple of ``(select_delta, cursor_delta)``.
        """
        if n >= (self.n_items - 1 - self.selected_item_index):
            self.selected_item_index = self.n_items - 1
            self.cursor_position = self.show_items_limit - 1
            return 0, 0
        select_delta, cursor_delta = 0, 0
        for _ in range(n):
            select_delta_, cursor_delta_ = self._press_down()
            select_delta += select_delta_
            cursor_delta += cursor_delta_
        return select_delta, cursor_delta

    def _press_up(self) -> T.Tuple[int, int]:
        if self.selected_item_index == 0:
            select_delta = 0
        else:
            self.selected_item_index -= 1
            select_delta = -1
        if self.cursor_position == 0:
            cursor_delta = 0
        else:
            self.cursor_position -= 1
            cursor_delta = -1
        return select_delta, cursor_delta

    def press_up(self, n: int = 1) -> T.Tuple[int, int]:
        """
        Move selector up to pick the previous item (if possible) in the dropdown menu.

        Example, before ``dropdown.press_up()``:

        .. code-block:: bash

            (Query):
            [ ] Beautiful is better than ugly.
                  subtitle 01
            [ ] Explicit is better than implicit.
                  subtitle 02
            [x] Simple is better than complex.
                  subtitle 03

        After:

        .. code-block:: bash

            (Query):
            [ ] Beautiful is better than ugly.
                  subtitle 01
            [x] Explicit is better than implicit.
                  subtitle 02
            [ ] Simple is better than complex.
                  subtitle 03

        :param n: move selector up ``n`` times.

        :return: tuple of ``(select_delta, cursor_delta)``.
        """
        if n >= self.selected_item_index:
            self.selected_item_index = 0
            self.cursor_position = 0
            return 0, 0
        select_delta, cursor_delta = 0, 0
        for _ in range(n):
            select_delta_, cursor_delta_ = self._press_up()
            select_delta += select_delta_
            cursor_delta += cursor_delta_
        return select_delta, cursor_delta

    def scroll_down(self, n: int = 1) -> T.Tuple[int, int]:
        """
        Scroll the dropdown menu down to show more items, also move the selector.

        :param n: scroll down ``n`` times. Each time we scroll down ``SCROLL_SPEED`` items.

        :return: tuple of ``(select_delta, cursor_delta)``.
        """
        return self.press_down(n * SCROLL_SPEED)

    def scroll_up(self, n: int = 1) -> T.Tuple[int, int]:
        """
        Scroll the dropdown menu up to show more items, also move the selector.

        :param n: scroll up ``n`` times. Each time we scroll up ``SCROLL_SPEED`` items.

        :return: tuple of ``(select_delta, cursor_delta)``.
        """
        return self.press_up(n * SCROLL_SPEED)

    @property
    def menu(self) -> T.List[T.Tuple[T_ITEM, bool]]:
        """
        The list of items to show in the UI. The list is determined by
        ``selected_item_index`` ``cursor_position`` and ``show_items_limit``
        together. It is a subset of ``self.items``.

        :return: a list of tuples, each tuple contains an item and a boolean value
            indicating whether the item is selected or not.
        """
        # this code is for debug only
        # print(self.n_items)
        # print(self.selected_item_index)
        # print(self.cursor_position)

        lower_index = self.selected_item_index - self.cursor_position
        upper_index = self.selected_item_index + (
            self.show_items_limit - self.cursor_position
        )
        menu = list()
        for ind, item in enumerate(self.items[lower_index:upper_index]):
            if ind == self.cursor_position:
                menu.append((item, True))
            else:
                menu.append((item, False))
        return menu



# -*- coding: utf-8 -*-

"""
Alfred Workflow UI simulator.
"""

import typing as T
import time
import subprocess
import dataclasses

import readchar

from .vendor.os_platform import IS_WINDOWS
from . import exc
from . import events
from .item import T_ITEM, Item
from .line_editor import LineEditor
from .dropdown import Dropdown
from .render import UIRender
from .debug import debugger


key_to_name = {
    readchar.key.CTRL_C: "CTRL_C",
    readchar.key.TAB: "TAB",
    readchar.key.CTRL_X: "CTRL_X",
    readchar.key.UP: "UP",
    readchar.key.DOWN: "DOWN",
    readchar.key.CTRL_E: "CTRL_E",
    readchar.key.CTRL_D: "CTRL_D",
    readchar.key.CTRL_R: "CTRL_R",
    readchar.key.CTRL_F: "CTRL_F",
    readchar.key.LEFT: "LEFT",
    readchar.key.RIGHT: "RIGHT",
    readchar.key.HOME: "HOME",
    readchar.key.END: "END",
    readchar.key.CTRL_H: "CTRL_H",
    readchar.key.CTRL_L: "CTRL_L",
    readchar.key.CTRL_G: "CTRL_G",
    readchar.key.CTRL_K: "CTRL_K",
    readchar.key.BACKSPACE: "BACKSPACE",
    readchar.key.DELETE: "DELETE",
    readchar.key.ENTER: "ENTER",
    readchar.key.CR: "CR",
    readchar.key.LF: "LF",
    readchar.key.CTRL_A: "CTRL_A",
    readchar.key.CTRL_W: "CTRL_W",
    readchar.key.CTRL_P: "CTRL_P",
}

if IS_WINDOWS:
    OPEN_CMD = "start"
else:
    OPEN_CMD = "open"


@dataclasses.dataclass
class DebugItem(Item):
    def enter_handler(self, ui: "UI"):
        subprocess.run([OPEN_CMD, str(debugger.path_log_txt)])


T_HANDLER = T.Callable[[str, T.Optional["UI"]], T.List[T_ITEM]]


class UI:
    """
    Zelfred terminal UI implementation.

    :param handler: a callable function that takes a query string as input and
        returns a list of items.
    :param hello_message: sometimes the handler execution for the first time
        run takes long. This message will be shown to user to indicate that
        the handler is still running, and the UI will show the returned items
        once the handler is done.
    :param capture_error: whether you want to capture the error and show it
        in the UI, or you want to raise the error immediately.
    :param process_input_immediately: the handler will be called immediately
        after user input. This is the default behavior.
    :param process_input_after_query_delay: the handler will be called after
        a delay of `query_delay` seconds after the first user input. For example,
        let's say the query delay is 0.3 second. At begin, the user input is empty,
        then user entered "a". If then the user entered "bcd" within 0.3 second,
        the handler will not be called to give you the latest items. The handler
        will be called until the user entered a new key, let's say "e", from
        0.4 second after he entered "a".
    :param query_delay: read above.
    """

    def __init__(
        self,
        handler: T_HANDLER,
        hello_message: T.Optional[str] = "Welcome to interactive UI!",
        capture_error: bool = True,
        process_input_immediately: bool = False,
        process_input_after_query_delay: bool = False,
        query_delay: float = 0.3,
    ):
        # -------------------- input arguments
        self.handler: T_HANDLER = handler
        self._handler_queue: T.List[T_HANDLER] = list()
        self.hello_message: T.Optional[str] = hello_message
        self.capture_error: bool = capture_error
        self.query_delay: float = query_delay

        # -------------------- internal implementation related
        self.render: UIRender = UIRender()
        self.event_generator = events.KeyEventGenerator()
        self._process_input: T.Callable
        flag = sum(
            [
                process_input_immediately,
                process_input_after_query_delay,
            ]
        )
        if flag == 0:
            process_input_immediately = True
        elif flag > 1:
            raise ValueError
        else: # pragma: no cover
            pass

        if process_input_immediately:
            self._process_input = self._process_input_v1_immediately
        elif process_input_after_query_delay:
            self._process_input = self._process_input_v2_after_query_delay
        else:  # pragma: no cover
            raise NotImplementedError

        # -------------------- items related --------------------
        self.line_editor: LineEditor = LineEditor()
        self.dropdown: Dropdown = Dropdown([])
        self.n_items_on_screen: int = 0

        # -------------------- controller flags --------------------
        self.need_run_handler: bool = True
        self.need_move_to_end: bool = True
        self.need_clear_items: bool = True
        self.need_clear_query: bool = True
        self.need_print_query: bool = True
        self.need_print_items: bool = True
        self.need_process_input: bool = True

    def _debug_controller_flags(self):
        print(f"self.need_run_handler = {self.need_run_handler}")
        print(f"self.need_move_to_end = {self.need_move_to_end}")
        print(f"self.need_clear_items = {self.need_clear_items}")
        print(f"self.need_clear_query = {self.need_clear_query}")
        print(f"self.need_print_query = {self.need_print_query}")
        print(f"self.need_print_items = {self.need_print_items}")
        print(f"self.need_process_input = {self.need_process_input}")

    def replace_handler(self, handler: T_HANDLER):
        """
        Replace the current handler with a new handler, and push the current
        handler to the handler queue (a last in first out stack).
        """
        self._handler_queue.append(self.handler)
        self.handler = handler

    def _clear_query(self):
        """
        Clear the ``(Query): {user_query}`` line.
        """
        self.render.clear_line_editor()
        return True

    def clear_query(self):
        """
        A wrapper of the ``_clear_query()`` method, ensures that the
        ``need_clear_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- clear_query --------------------")
        try:
            if self.need_clear_query:
                debugger.log("before clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._clear_query()
                debugger.log("after clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_clear_query = True

    def _clear_items(self) -> bool:
        """
        Clear the item dropdown menu.
        """
        if self.render.line_number > 1:
            # time.sleep(1)
            self.render.clear_dropdown()
            # time.sleep(1)
            return True
        else:
            return False

    def clear_items(self):
        """
        A wrapper of the ``_clear_items()`` method, ensures that the
        ``need_clear_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- clear_items --------------------")
        try:
            if self.need_clear_items:
                debugger.log("before clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                flag = self._clear_items()
                debugger.log("after clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                if flag:
                    debugger.log(f"nothing happen")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_clear_items = True

    def _print_query(self):
        """
        Print the ``(Query): {user_query}`` line
        """
        content = self.render.print_line_editor(self.line_editor)
        debugger.log(f"printed: {content!r}")

    def print_query(self):
        """
        A wrapper of the core logic for printing query, ensures that the
        ``need_print_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- print_query --------------------")
        try:
            if self.need_print_query:
                debugger.log("before print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._print_query()
                debugger.log("after print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_print_query = True

    def _print_items(self):
        """
        Core logic for printing items in the dropdown menu.
        """
        debugger.log("render dropdown")
        n_items_on_screen = self.render.print_dropdown(
            self.dropdown, self.render.terminal.width
        )
        self.n_items_on_screen = n_items_on_screen

    def print_items(self):
        """
        A wrapper of the core logic for printint items, ensures that the
        ``need_print_items`` and ``need_run_handler`` flag is set to ``True``
        at the end regardless of whether an exception is raised.
        """
        debugger.log("-------------------- print_items --------------------")
        try:
            if self.need_print_items:
                debugger.log("before print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._print_items()
                # manually move the cursor to the edge to maximize the room for rendering
                self.render.move_down(1)
                self.render.move_up(1)
                debugger.log("after print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            debugger.log(f"move_cursor_to_line_editor ...")
            debugger.log(f"render.line_number: {self.render.line_number}")
            debugger.log(f"render.n_lines: {self.render.n_lines}")
            # always move cursor back to line editor after printing items
            n_vertical, n_horizontal = self.render.move_cursor_to_line_editor(
                self.line_editor
            )
            debugger.log(f"n_vertical: {n_vertical}")
            debugger.log(f"n_horizontal: {n_horizontal}")
            self.need_print_items = True
            self.need_run_handler = True

    def process_key_pressed_input(self, pressed: str):
        """
        Process user keyboard input.

        - UP: move up
        - DOWN: move down
        - LEFT: move left
        - RIGHT: move right
        - HOME: move to start of the line
        - END: move to end of the line

        - CTRL + E: move up
        - CTRL + D: move down
        - CTRL + R: scroll up
        - CTRL + F: scroll down
        - CTRL + H: move left
        - CTRL + L: move right
        - CTRL + G: move word left
        - CTRL + K: move word right

        - CTRL + X: clear input

        Actions:

        - Enter:
        - CTRL + A:
        - CTRL + W:
        - CTRL + P:
        """
        pressed_key_name = key_to_name.get(pressed, pressed)
        debugger.log(f"pressed: {pressed_key_name!r}, key code: {pressed!r}")

        if pressed == readchar.key.CTRL_C:
            raise KeyboardInterrupt()

        if pressed == readchar.key.TAB:
            self.line_editor.clear_line()
            selected_item = self.dropdown.selected_item
            if selected_item.autocomplete:
                self.line_editor.enter_text(selected_item.autocomplete)
            return

        if pressed == readchar.key.CTRL_X:
            self.line_editor.clear_line()
            return

        if pressed in (
            readchar.key.UP,
            readchar.key.DOWN,
            readchar.key.CTRL_E,
            readchar.key.CTRL_D,
            readchar.key.CTRL_R,
            readchar.key.CTRL_F,
        ):
            self.need_run_handler: bool = False
            self.need_clear_query: bool = False
            self.need_print_query: bool = False

            if pressed in (readchar.key.UP, readchar.key.CTRL_E):
                self.dropdown.press_up()
            elif pressed in (readchar.key.DOWN, readchar.key.CTRL_D):
                self.dropdown.press_down()
            elif pressed == readchar.key.CTRL_R:
                self.dropdown.scroll_up()
            elif pressed == readchar.key.CTRL_F:
                self.dropdown.scroll_down()
            else:  # pragma: no cover
                raise NotImplementedError
            return

        # note: on windows terminal, the backspace and CTRL+H key code are the same
        # we have to sacrifice the CTRL+H key to keep BACKSPACE working,
        # so we put this code block before CTRL+H
        if pressed == readchar.key.BACKSPACE:
            self.line_editor.press_backspace()
            return

        if pressed == readchar.key.DELETE:
            self.line_editor.press_delete()
            return

        if pressed in (
            readchar.key.LEFT,
            readchar.key.RIGHT,
            readchar.key.HOME,
            readchar.key.END,
            readchar.key.CTRL_H,  # note, CTRL+H won't work on Windows
            readchar.key.CTRL_L,
            readchar.key.CTRL_G,
            readchar.key.CTRL_K,
        ):
            self.need_run_handler = False
            self.need_move_to_end = False
            self.need_clear_items = False
            self.need_clear_query = False
            self.need_print_query = False
            self.need_print_items = False
            if pressed in (readchar.key.LEFT, readchar.key.CTRL_H):
                self.line_editor.press_left()
            elif pressed in (readchar.key.RIGHT, readchar.key.CTRL_L):
                self.line_editor.press_right()
            elif pressed == readchar.key.HOME:
                self.line_editor.press_home()
            elif pressed == readchar.key.END:
                self.line_editor.press_end()
            elif pressed == readchar.key.CTRL_G:
                self.line_editor.move_word_backward()
            elif pressed == readchar.key.CTRL_K:
                self.line_editor.move_word_forward()
            else:  # pragma: no cover
                raise NotImplementedError
            return

        if pressed in (
            readchar.key.ENTER,
            readchar.key.CR,
            readchar.key.LF,
            readchar.key.CTRL_A,
            readchar.key.CTRL_W,
            readchar.key.CTRL_P,
        ):
            if self.dropdown.n_items == 0:
                raise exc.EndOfInputError(
                    selection="select nothing",
                )
            else:
                self.move_to_end()
                if self.dropdown.items:
                    selected_item = self.dropdown.selected_item
                    if pressed in (
                        readchar.key.ENTER,
                        readchar.key.CR,
                        readchar.key.LF,
                    ):
                        selected_item.enter_handler(ui=self)
                    elif pressed == readchar.key.CTRL_A:
                        selected_item.ctrl_a_handler(ui=self)
                    elif pressed == readchar.key.CTRL_W:
                        selected_item.ctrl_w_handler(ui=self)
                    elif pressed == readchar.key.CTRL_P:
                        selected_item.ctrl_p_handler(ui=self)
                    else:  # pragma: no cover
                        raise NotImplementedError
                raise exc.EndOfInputError(selection=selected_item)

        if pressed == readchar.key.F1:
            raise exc.JumpOutLoopError

        self.line_editor.press_key(pressed)

    def _process_input_v1_immediately(self):
        event = self.event_generator.next()
        if isinstance(event, events.KeyPressedEvent):
            self.process_key_pressed_input(pressed=event.value)

    def _process_input_v2_after_query_delay(self):
        start_time: T.Optional[float] = None
        while 1:
            event = self.event_generator.next()
            if isinstance(event, events.KeyPressedEvent):
                self.process_key_pressed_input(pressed=event.value)
            self.clear_query()
            self.print_query()
            self.render.move_cursor_to_line_editor(self.line_editor)
            if start_time is None:
                start_time = time.time()
            else:
                event_time = time.time()
                if event_time - start_time > self.query_delay:
                    break

    def process_input(self):
        """
        A wrapper of the core logic for processing input, ensures that the
        ``need_process_input`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- process_input --------------------")
        try:
            if self.need_process_input:
                self._process_input()
        finally:
            self.need_process_input = True

    def run_handler(self, items: T.Optional[T.List[T_ITEM]] = None):
        if self.need_run_handler:
            debugger.log("need to run handler")
            if items is None:
                debugger.log("run handler")
                items = self.handler(self.line_editor.line, self)
            else:
                debugger.log("explicitly give items, skip handler")
            self.dropdown.update(items)

    def move_to_end(self):
        """
        Move the cursor to the next line of the end of the dropdown menu.
        """
        debugger.log("-------------------- move_to_end --------------------")
        try:
            if self.need_move_to_end:
                debugger.log("need to move to end")
                debugger.log("before move to end")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                n_down = self.render.move_to_end()
                debugger.log(f"move down {n_down} lines")
                debugger.log("after move to end")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_move_to_end = True

    def print_hello_items(self):
        items = [
            Item(
                uid="uid-hello-item",
                title="Welcome to the interactive UI",
            )
        ]
        self.dropdown.update(items)
        self.print_items()

    def print_debug_items(self, e: Exception):
        # if dropdown has items, then there must be one selected item
        # include the selected item information in the error message
        if self.dropdown.items:
            selected_item = self.dropdown.selected_item
            title = selected_item.title
            # handle the error in debug loop special case
            if title.startswith("Error on item: "):
                title = title[len("Error on item: ") :]
            processed_title = self.render.process_title(
                title,
                self.render.terminal.width - 15,
            )
            items = [
                DebugItem(
                    uid="uid",
                    title=f"Error on item: {processed_title}",
                    subtitle=f"{e!r}",
                )
            ]
        else:
            items = [
                DebugItem(
                    uid="uid",
                    title=f"Error on item: NA",
                    subtitle=f"{e!r}",
                )
            ]
        self.dropdown.update(items)
        self.print_items()

    def initialize_loop(self):
        debugger.log("=== initialize loop start ===")
        self.print_query() # show initial query, mostly it is empty
        self.print_hello_items() # show hello items
        self.run_handler() # run handler using initial query
        self.move_to_end() # move cursor to the end
        self.clear_items() # clear items
        self.clear_query() # clear query
        self.print_query() # print query
        self.print_items() # print items
        debugger.log("=== initialize loop end ===")

    def main_loop(self, _ith: int=0):
        while True:
            _ith += 1
            debugger.log(f"=== {_ith}th main loop start ===")
            self.process_input()
            # self._debug_controller_flags()
            self.run_handler()
            self.move_to_end()
            self.clear_items()
            self.clear_query()
            self.print_query()
            self.print_items()
            debugger.log(f"=== {_ith}th main loop end ===")

    def debug_loop(self, e: Exception):
        """
        Display error message in the UI and wait for new user input.
        New user input may fix the problem or trigger another error.
        """
        debugger.log("=== debug loop start ===")
        self.move_to_end()
        self.clear_items()
        self.clear_query()
        self.print_query()
        self.print_debug_items(e)
        debugger.log("=== debug loop end ===")

    def run(self, _do_init: bool = True):
        """
        Run the UI.

        Read :ref:`ui-event-loop`
        (or `this link <https://zelfred.readthedocs.io/en/latest/03-UI-Event-Loop/index.html>`_)
        for more information.
        """
        try:
            if _do_init:
                self.initialize_loop()
            self.main_loop()
        except exc.EndOfInputError as e:
            return e.selection
        except exc.JumpOutLoopError:
            if self._handler_queue:
                self.handler = self._handler_queue.pop()
            self.line_editor.clear_line()
            self.run_handler()
            self.move_to_end()
            self.clear_items()
            self.clear_query()
            self.print_query()
            self.print_items()
            return self.run(_do_init=False)
        except KeyboardInterrupt:
            self.move_to_end()
            print("🔴 keyboard interrupt, exit.", end="")
        except Exception as e:
            if self.capture_error:
                self.debug_loop(e)
                return self.run(_do_init=False)
            else:
                raise e
        finally:
            print("")



# -*- coding: utf-8 -*-

"""
The user input in the line editor is called :class:`Query`.
This module provides an objective oriented interface to parse user input string
into machine friendly :class:`Query` object.
"""

import typing as T
import dataclasses


_SEP = "____"


class QueryParser:
    """
    Utility class that can parse string to :class:`Query`. Naturally, it is
    just a tokenizer.

    :param delimiter: the delimiter to split the query string.
    """

    def __init__(
        self,
        delimiter: T.Union[str, T.List[str]] = " ",
    ):
        if isinstance(delimiter, str):
            self.delimiter = [
                delimiter,
            ]
        else:
            self.delimiter = delimiter

    def parse(self, s: str) -> "Query":
        """
        Convert string query to structured :class:`Query` object.

        :param s: a string.
        """
        for sep in self.delimiter:
            s = s.replace(sep, _SEP)
        parts = s.split(_SEP)
        trimmed_parts = [c.strip() for c in parts if c.strip()]
        return Query(
            raw=s,
            parts=parts,
            trimmed_parts=trimmed_parts,
        )


DEFAULT_QUERY_PARSER = QueryParser()


@dataclasses.dataclass
class Query:
    """
    Structured query object. This is very useful to parse the input of UI handler.

    :param parts: the parts of query string split by delimiter. For example,
        if the user input is ``"hello world!"``, then ``parts`` is
        ``["hello", "world!"]``.
    :param trimmed_parts: similar to parts, but each part is white-space stripped
        For example, if the user input is ``" hello world "``, then ``parts`` is
        ``["", "hello", "world", ""]``, and ``trimmed_parts`` is ``["hello", "world"]``.

    Usage::

        >>> q = Query.from_str("  a   b   c  ")
        >>> q.trimmed_parts
        ['a', 'b', 'c']
        >>> q.n_trimmed_parts
        3
    """

    raw: str = dataclasses.field()
    parts: T.List[str] = dataclasses.field()
    trimmed_parts: T.List[str] = dataclasses.field()

    @classmethod
    def from_str(cls, s: str, parser=DEFAULT_QUERY_PARSER):
        """
        Parse query from string using the given parser.
        """
        return parser.parse(s)

    @property
    def n_parts(self) -> int:
        """
        The number of items in the ``parts``.
        """
        return len(self.parts)

    @property
    def n_trimmed_parts(self) -> int:
        """
        The number of items in the ``trimmed_parts``.
        """
        return len(self.trimmed_parts)



# -*- coding: utf-8 -*-

"""
This module implements the render logic of the terminal UI.
"""

import typing as T
import dataclasses
import sys

from blessed import Terminal

from .item import T_ITEM
from .line_editor import LineEditor
from .dropdown import Dropdown
from .exc import TerminalTooSmallError


@dataclasses.dataclass
class Render:
    """
    Generic UI render. It can print string, print line, clear line, and move
    cursor up and down.

    :param terminal: blessed terminal object.
    :param line_number: store the current line number of the cursor.
        if 0, means it is at the first line.
        if 1, means it is at the second line.
    :param n_lines: store the total number of printed lines in the terminal.
    """

    terminal: Terminal = dataclasses.field(default_factory=Terminal)
    line_number: int = dataclasses.field(default=0)
    n_lines: int = dataclasses.field(default=0)

    def __post_init__(self):
        pass

    def _debug(self):
        print(f"line_number = {self.line_number}")
        print(f"n_lines = {self.n_lines}")

    def print_str(
        self,
        str_tpl: str,
        new_line=False,
        **kwargs,
    ) -> str:
        """
        打印一个字符串, 可以选择是否换行. 并且自动更新 line_number 的值.

        :param str_tpl: string template, 一个字符串模板.
        :param new_line: 如果是 True, 那么会自动换行, 否则不会. 默认不会.
        :param kwargs: 额外的传递给 str_tpl 的参数.
        """
        if new_line:
            self.line_number += 1
            self.n_lines += 1
        content = str_tpl.format(**kwargs)
        print(content, end="\n" if new_line else "")
        sys.stdout.flush()
        return content

    def print_line(
        self,
        str_tpl: str,
        new_line: bool = True,
        **kwargs,
    ) -> str:
        """
        打印一行, 默认自动换行.

        :param str_tpl: string template, 一个字符串模板.
        :param new_line: 如果是 True, 那么会自动换行, 否则不会. 默认自动换行.
        :param kwargs: 额外的传递给 str_tpl 的参数.
        """
        return self.print_str(
            str_tpl + self.terminal.clear_eol(),
            new_line=new_line,
            **kwargs,
        )

    def _force_initial_column(self) -> str:
        """
        用回车符 (注意! 不是换行符) 把光标移动到本行初始位置.
        """
        return self.print_str("\r")

    def move_to_start(self) -> int:
        """
        把光标移动到初始位置. 本质上是把光标向上回退移动到第一行, 然后再用回车符把光标移动到本行初始位置.
        """
        n = self.line_number
        if n:
            print(n * self.terminal.move_up, end="")
        print("\r", end="")
        sys.stdout.flush()
        self.line_number = 0
        return n

    def move_up(self, n: int):
        """
        把光标移动到初始位置. 本质上是把光标向上回退移动到第一行, 然后再用回车符把光标移动到本行初始位置.
        """
        print(n * self.terminal.move_up, end="")
        sys.stdout.flush()
        self.line_number -= n

    def move_down(self, n: int):
        """
        把光标移动到初始位置. 本质上是把光标向上回退移动到第一行, 然后再用回车符把光标移动到本行初始位置.
        """
        print(n * self.terminal.move_down, end="")
        sys.stdout.flush()
        self.line_number += n

    def clear_n_lines(self, n: int):
        """
        把光标以上的 n 行清空, 并把光标移动到行首. 常用于清除掉已经打印过的内容.
        """
        if n > self.line_number:
            raise ValueError
        for _ in range(n):
            print(self.terminal.move_up, end="")
            print(self.terminal.clear_eol(), end="")
            print(self.terminal.clear_bol(), end="")
        self._force_initial_column()
        self.line_number -= n
        self.n_lines -= n

    def clear_all(self):
        """
        清除所有内容, 并把光标移动到行首.

        Clear
        """
        self.clear_n_lines(n=self.line_number)

    @property
    def width(self):
        return self.terminal.width or 80

    @property
    def height(self):
        return self.terminal.width or 24


@dataclasses.dataclass
class UIRender(Render):
    """
    The Terminal UI Render. It extends the :class:`Render` class and highly
    optimized for the ``zelfred`` terminal UI layout.

    See below example, ``|`` represents the cursor:

    .. code-block::

        (Query): user query here|
        [x] item 1 title here
              item 1 subtitle here
        [ ] item 2 title here
              item 2 subtitle here
        [ ] item 3 title here
              item 3 subtitle here

    The first line ``(Query): user query here|`` is the user input box, it always
    starts with ``(Query): ``, and user can enter any text input after that.
    The cursor cannot go beyond the ``: `` part.

    User can use ``Left``, ``Right``, ``Backspace`` and ``Delete`` keys to edit the
    user input box.

    Below the first line is the items drop down menu. Each item has two lines.

    The first line is the title, it always starts with ``[x] `` or ``[ ] ``.
    ``[x] `` means the item is selected, ``[ ] `` means the item is not selected.
    You can only select one item at a time, and by default the first item is selected.
    There always be one item selected.

    The second line is the subtitle, it has 2 indent spaces comparing to the title.

    User can use the ``UP`` and ``DOWN`` keys to navigate the items drop down menu.

    The dropdown menu can show up to 10 items at a time, if the dropdown menu
    has more than 10 items, user can scroll down to see the rest of the items using
    the ``DOWN`` key. The ``CTRL + UP`` and ``CTRL + DOWN`` key can scroll up and down
    10 items at a time.
    """

    prompt: str = dataclasses.field(default="(Query)")
    checked_mark: str = dataclasses.field(default="[x]")
    not_checked_mark: str = dataclasses.field(default="[ ]")
    subtitle_pad: str = dataclasses.field(default=" " * 6)

    def __post_init__(self):
        super().__post_init__()

    # --------------------------------------------------------------------------
    # line editor
    # --------------------------------------------------------------------------
    def print_line_editor(self, line_editor: LineEditor) -> str:
        """
        Render the line editor, the ``(Query): user query here|`` part. And move
        the cursor to the beginning of the next line, so it's ready to print the
        dropdown menu.

        It assumes that there is nothing in the terminal UI before running this.

        :param line_editor: the new :class:`~zelfred.line_editor.LineEditor` object.

        :return: text of the line editor.
        """

        return self.print_line(
            "{t.bold}{t.cyan}{prompt}: {t.normal}{line_editor.line}",
            prompt=self.prompt,
            line_editor=line_editor,
            t=self.terminal,
        )

    def clear_line_editor(self):
        """
        Clear the line editor (the query input box at the first line).
        It doesn't require to move the cursor to the beginning of the line editor
        before calling this function, it will handle that automatically.
        This function will move the cursor to the beginning of the line editor
        at the end.

        Before::

            (Query): user query here|
            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here

        Before::

            |   # <- this line is empty
            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here
        """
        self.move_to_start()
        self.print_str(self.terminal.clear_eol(), end="")

    def update_line_editor(self, line_editor: LineEditor) -> str:
        """
        Replace the user input with the new line editor, and move the cursor
        to the beginning of the next line, so it's ready to print the
        dropdown menu.

        :param line_editor: the new :class:`~zelfred.line_editor.LineEditor` object.

        :return: text of the line editor.
        """
        self.clear_line_editor()
        return self.print_line_editor(line_editor)

    # --------------------------------------------------------------------------
    # dropdown
    # --------------------------------------------------------------------------
    def process_title(self, title: str, line_width: int):
        """
        Make sure the title fix the width of the terminal UI.

        :param title: the item title.
        :param line_width: the max width of the terminal UI, it will truncate
            the title and subtitle if they are too long.
        """
        space = line_width - 4 - 1
        if len(title) > space:
            half = (space - 3) // 2
            return title[:half] + "..." + title[-half:]
        else:
            return title

    def process_subtitle(self, subtitle: str, line_width: int):
        """
        Make sure the subtitle fix the width of the terminal UI.

        :param subtitle: the item subtitle.
        :param line_width: the max width of the terminal UI, it will truncate
            the title and subtitle if they are too long.
        """
        space = line_width - 6 - 1
        if len(subtitle) > space:
            half = (space - 3) // 2
            return subtitle[:half] + "..." + subtitle[-half:]
        else:
            return subtitle

    def print_item(self, item: T_ITEM, selected: bool, line_width: int):
        """
        Render one item in the dropdown menu. It looks like::

            [x] item 1 title here
                - item 1 subtitle here

        :param item: the :class:`~zelfred.item.Item` object to render.
        :param line_width: the max width of the terminal UI, it will truncate
            the title and subtitle if they are too long.
        """
        if selected:
            color = self.terminal.cyan
            symbol = self.checked_mark
        else:
            color = self.terminal.normal
            symbol = self.not_checked_mark

        self.print_line(
            "{t.bold}{color}{symbol} {color}{title}{t.normal}",
            color=color,
            symbol=symbol,
            title=self.process_title(item.title_text, line_width),
            t=self.terminal,
        )
        self.print_line(
            "{pad}{t.normal}{subtitle}",
            pad=self.subtitle_pad,
            subtitle=self.process_subtitle(item.subtitle_text, line_width),
            t=self.terminal,
        )

    def print_dropdown(self, dropdown: Dropdown, line_width: int) -> int:
        """
        Render all items in the dropdown menu, it looks like::

            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here

        It assumes that the terminal UI is showing the query line editor and
        there's no item in the dropdown menu in the terminal UI before running this.

        :param dropdown: the :class:`~zelfred.dropdown.Dropdown` object to render.
        :param line_width: the max width of the terminal UI, it will truncate
            the title and subtitle if they are too long.

        :return: number of items rendered.
        """
        # the current terminal height may not be able to fit all items,
        # so we may need to update the ``self.dropdown._show_items_limit``
        # to fit the terminal height.
        terminal_height = self.terminal.height
        if terminal_height <= 9:
            raise TerminalTooSmallError(
                "Terminal height is too small to render the UI! "
                "It has to have at least 8 lines."
            )
        terminal_items_limit = (terminal_height - 3) // 2
        dropdown.show_items_limit = min(
            dropdown.show_items_limit,
            terminal_items_limit,
        )

        # raise error if the terminal width is too small
        terminal_width = self.terminal.width
        final_line_width = min(terminal_width, line_width)
        if terminal_width < 80:
            raise TerminalTooSmallError(
                "Terminal width is too small to render the UI! "
                "It has to have at least 80 ascii character wide."
            )

        # if the cursor is at the first line, move to beginning of the second line
        # before printing
        if self.line_number == 0:
            print(self.terminal.move_down, end="")
            print("\r", end="")
            sys.stdout.flush()

        # print the dropdown menu
        menu = dropdown.menu
        for item, selected in dropdown.menu:
            self.print_item(item, selected=selected, line_width=final_line_width)
        n_item = len(menu)
        return n_item

    def clear_dropdown(self):
        """
        Before::

            (Query): user query here|
            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here

        Before::

            (Query): user query here
            |
        """
        self.move_to_end()
        # self._debug()
        if self.n_lines > 1:
            self.clear_n_lines(n=self.line_number - 1)

    def update_dropdown(self, dropdown: Dropdown, line_width: int) -> int:
        """
        Replace the user input with the new line editor, and move the cursor
        to the beginning of the next line, so it's ready to print the
        dropdown menu.

        :param dropdown: the :class:`~zelfred.dropdown.Dropdown` object to render.
        :param line_width: the max width of the terminal UI, it will truncate
            the title and subtitle if they are too long.

        :return: number of items rendered.
        """
        self.clear_dropdown()
        return self.print_dropdown(dropdown, line_width)

    def move_cursor_to_line_editor(self, line_editor: LineEditor) -> T.Tuple[int, int]:
        """
        After the :meth:`Dropdown.print_dropdown` is called, the cursor is at the
        end of the UI. This method moves the cursor back to the user input box,
        so user can keep typing.

        Here's an example::

            (Query): user query here| <- want to move to here
            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here
            | <- cursor is currently here

        :return: n_vertical is the number of line to move up, n_horizontal is
            the number of character to move right.
        """
        n_vertical = self.move_to_start()
        n_horizontal = len(self.prompt) + 2 + line_editor.cursor_position
        self.print_str(self.terminal.move_right(n_horizontal), end="")
        return n_vertical, n_horizontal

    def print_ui(self, line_editor: LineEditor, dropdown: Dropdown) -> int:
        """
        Render the entire UI, and move the cursor to the right position.

        It assumes the terminal has nothing.
        """
        self.print_line_editor(line_editor)
        n_items = self.print_dropdown(dropdown, line_width=self.terminal.width)
        self.move_cursor_to_line_editor(line_editor)
        return n_items

    def move_to_end(self) -> int:
        """
        Move the cursor to the end, this method will be used before exit.
        Here's an example::

            (Query): user query here| <- cursor is currently here
            [x] item 1 title here
                  item 1 subtitle here
            [ ] item 2 title here
                  item 2 subtitle here
            [ ] item 3 title here
                  item 3 subtitle here
            | <- want to move to here

        :param n_items: number of items in the dropdown menu.
        """
        move_down_n_lines = self.n_lines - self.line_number
        if move_down_n_lines:
            self.print_str(move_down_n_lines * self.terminal.move_down, end="")
            self.line_number += move_down_n_lines
        return move_down_n_lines

    def clear_ui(self):
        """
        Clear the entire UI, and move the cursor to the right position.
        """
        self.clear_dropdown()
        self.clear_line_editor()



# -*- coding: utf-8 -*-

"""
Usage example:

.. code-block:: python

    from fixa.os_platform import IS_WINDOWS, IS_MACOS, IS_LINUX

Reference:

- https://docs.python.org/3/library/sys.html#sys.platform
"""

import sys

IS_WINDOWS = False
IS_MACOS = False
IS_LINUX = False

_platform = sys.platform
if _platform in ["win32", "cygwin"]:
    IS_WINDOWS = True
elif _platform == "darwin":
    IS_MACOS = True
elif _platform == "linux":
    IS_LINUX = True
else:
    pass



# -*- coding: utf-8 -*-

"""
This module is used to debug ``zelfred`` library.
"""

from datetime import datetime
from .paths import path_log_txt


class Debugger:
    """
    This class is used to debug ``zelfred`` library.

    :param path_log_txt: the path to the log file.
    """
    def __init__(self):
        self.path_log_txt = path_log_txt
        self._enable = False

    def reset(self):
        """
        Remove the log file.
        """
        try:
            self.path_log_txt.unlink()
        except FileNotFoundError:
            pass

    def enable(self):
        """
        Enable the debugger.
        """
        self._enable = True

    def disable(self):
        """
        Disable the debugger.
        """
        self._enable = False

    def _log(self, text: str):
        with self.path_log_txt.open("a") as f:
            ts = datetime.now().strftime("[%Y-%m-%d %H:%M:%S]")
            line = f"{ts} {text}\n"
            f.write(line)

    def log(self, text: str):
        """
        Log (append a new line) a text to the log file.
        """
        if self._enable:
            self._log(text)


debugger = Debugger()



# -*- coding: utf-8 -*-

"""
Line editor implementation.
"""

SEP_LIST = "!@#$%^&*()-_+={[}]|\\:;"'<,>.?/'


class LineEditor:
    """
    Simulate a user input line editor. User can type characters, move cursor,
    backspace, delete, clear line, etc ...

    For example, the ``(Query): beautiful|`` (``|`` is the cursor) in the
    following UI is the line editor.

    .. code-block:: bash

        (Query): beautiful|
        [x] Beautiful is better than ugly.
              subtitle 01
        [ ] Explicit is better than implicit.
              subtitle 02
        [ ] Simple is better than complex.
              subtitle 03

    Empty line editor:

    .. code-block:: bash

        |

    User entered some text, and the cursor is at the end:

    .. code-block:: bash

        my text|

    User entered some text, and the cursor is in the middle:

    .. code-block:: bash

        my |text

    :param chars: a list of characters, representing the current line.
        For example, if ``chars = ["m", "y", " ", "t", "e", "x", "t"]``,
        then the current line is ``my text``.
    :param cursor_position: the current cursor position. 0 means the cursor is
        at the beginning of the line. 1 means it is after the first character.
        when cursor_position == len(chars), it means the cursor is at the end.
        For example, if the text is ``my text`` and the ``cursor_position = 5``
        Then the cursor is at ``my te|xt``.
    """

    def __init__(self):
        self.chars = []
        self.cursor_position = 0

    def is_cursor_at_begin_of_line(self) -> bool:
        """
        Check if the cursor is at the beginning of the line.
        """
        return self.cursor_position == 0

    def is_cursor_at_end_of_line(self) -> bool:
        """
        Check if the cursor is at the end of the line.
        """
        return self.cursor_position == len(self.chars)

    def enter_text(self, text: str):
        """
        Enter text into the line editor.
        """
        for char in text:
            self.press_key(key=char)

    def _press_key(self, key: str):
        if self.is_cursor_at_end_of_line():
            self.chars.append(key)
            self.cursor_position += 1
        else:
            self.chars.insert(self.cursor_position, key)
            self.cursor_position += 1

    def press_key(self, key: str, n: int = 1):
        """
        Enter a key into the line editor. Also move cursor to the right.

        :param key: the entered character of the key.
        :param n: number of times to enter the key.
        """
        for _ in range(n):
            self._press_key(key)

    def _press_backspace(self):
        if self.cursor_position == 0:
            pass
        elif self.cursor_position == len(self.chars):
            self.chars.pop()
            self.cursor_position -= 1
        else:
            self.cursor_position -= 1
            self.chars.pop(self.cursor_position)

    def press_backspace(self, n: int = 1):
        """
        Delete character backwards in the line editor. Also move cursor to the left.

        :param n: number of characters to delete.
        """
        for _ in range(n):
            self._press_backspace()

    def _press_left(self):
        if self.cursor_position != 0:
            self.cursor_position -= 1

    def press_left(self, n: int = 1):
        """
        Move cursor to the left.

        :param n: number of times to move cursor to the left.
        """
        for _ in range(n):
            self._press_left()

    def press_home(self):
        """
        Move cursor to the beginning of the line.
        """
        self.cursor_position = 0

    def _press_delete(self):
        if self.cursor_position == len(self.chars):
            pass
        else:
            self.chars.pop(self.cursor_position)

    def press_delete(self, n: int = 1):
        """
        Delete character forwards in the line editor. Also the cursor stays.

        :param n: number of characters to delete.
        """
        for _ in range(n):
            self._press_delete()

    def _press_right(self):
        if self.cursor_position != len(self.chars):
            self.cursor_position += 1

    def press_right(self, n: int = 1):
        """
        Move cursor to the right.

        :param n: number of times to move cursor to the right.
        """
        for _ in range(n):
            self._press_right()

    def press_end(self):
        """
        Move cursor to the end of the line.
        """
        self.cursor_position = len(self.chars)

    def clear_line(self):
        """
        Delete all user inputs and move cursor to the beginning of the line.
        """
        self.chars.clear()
        self.cursor_position = 0

    def clear_backward(self):
        """
        Delete all user inputs before the cursor and move cursor to the
        beginning of the line.
        """
        self.chars = self.chars[self.cursor_position:]
        self.cursor_position = 0

    def clear_forward(self):
        """
        Delete all user inputs after the cursor and the cursor stays.
        """
        self.chars = self.chars[: self.cursor_position]
        self.cursor_position = len(self.chars)

    def replace_text(self, text: str):
        """
        Replace all user inputs with the given text.

        :param text: the text to replace with.
        """
        self.clear_line()
        self.enter_text(text)

    move_to_start = press_home
    move_to_end = press_end

    def move_word_backward(self):
        """
        Move cursor to the beginning of the previous word.
        """
        # 先获得光标之前的字符串
        line = self.value
        # 按照空格分割开, words 里面的元素可以是空字符串
        words = line.split(" ")
        # print(f"before: words = {words}")
        # 从后往前找到第一个非空字符串的 index
        ind = None
        for i, word in enumerate(words[::-1]):
            if word:
                ind = i
                break
        # print(f"ind = {ind}")
        # 如果找到了非空字符串
        if ind is not None:
            # 那么保留所有非空字符串之前的 word, 并把最后一个非空字符串替换成空字符串
            # 这样即可算的 cursor position
            if ind:
                words = words[:-ind]
            words[-1] = ""
            # print(f"after: words = {words}")
            self.cursor_position = len(" ".join(words))
        # 如果找不到非空字符串, 则移动到行首
        else:
            self.cursor_position = 0

    def move_word_forward(self):
        """
        Move cursor to the beginning of the next word.
        """
        # 先获得光标之后的字符串
        line = "".join(self.chars[self.cursor_position:])
        # 按照空格分割开, words 里面的元素可以是空字符串
        words = line.split(" ")
        # print(f"before: words = {words}")
        # 从前往后找到第一个非空字符串
        ind = None
        for i, word in enumerate(words):
            if word:
                ind = i
                break
        # print(f"ind = {ind}")
        # 如果找到了非空字符串, 则计算这个非空字符串起之前的所有字符串的总长度
        # 这个长度就是 cursor 要移动的距离
        if ind is not None:
            words = words[:(ind+1)]
            # print(f"after: words = {words}")
            self.cursor_position += len(" ".join(words))
        # 如果找不到非空字符串, 则移动到行尾
        else:
            self.cursor_position = len(self.chars)

    @property
    def line(self) -> str:
        """
        Return the displayed line.

        Example: ``ali|ce`` -> line = alice
        """
        return "".join(self.chars)

    @property
    def value(self) -> str:
        """
        The value of the user input, it is the text before the cursor.

        Example: ``ali|ce`` -> value = ali
        """
        return "".join(self.chars[: self.cursor_position])



# -*- coding: utf-8 -*-

from zelfred.line_editor import LineEditor


def _test_is_cursor_at_begin_or_end_of_line():
    le = LineEditor()
    assert le.is_cursor_at_begin_of_line() is True
    assert le.is_cursor_at_end_of_line() is True

    le.press_key("a")
    assert le.is_cursor_at_begin_of_line() is False
    assert le.is_cursor_at_end_of_line() is True

    le.press_left()
    assert le.is_cursor_at_begin_of_line() is True
    assert le.is_cursor_at_end_of_line() is False


def _test_press_key():
    le = LineEditor()
    le.press_key("a")
    assert le.line == "a"
    assert le.value == "a"


def _test_enter_text():
    le = LineEditor()
    le.enter_text("abc")
    assert le.line == "abc"
    assert le.value == "abc"


def _test_press_backspace():
    le = LineEditor()
    le.enter_text("abc")
    le.press_backspace()
    assert le.line == "ab"
    assert le.value == "ab"

    le.press_backspace(2)
    assert le.line == ""
    assert le.value == ""

    le.press_backspace(5)
    assert le.line == ""
    assert le.value == ""


def _test_press_left():
    le = LineEditor()
    le.enter_text("alice")

    le.press_left()
    assert le.line == "alice"
    assert le.value == "alic"

    le.press_backspace()
    assert le.line == "alie"
    assert le.value == "ali"

    le.press_left()
    assert le.line == "alie"
    assert le.value == "al"

    le.press_key("X")
    assert le.line == "alXie"
    assert le.value == "alX"


def _test_press_home():
    le = LineEditor()
    le.enter_text("alice")

    le.press_home()
    assert le.line == "alice"
    assert le.value == ""

    le.press_left()
    assert le.line == "alice"
    assert le.value == ""


def _test_press_delete():
    le = LineEditor()
    le.enter_text("abcde")
    le.press_delete()
    assert le.line == "abcde"
    assert le.value == "abcde"

    le.press_left(2)

    le.press_delete()
    assert le.line == "abce"
    assert le.value == "abc"

    le.press_delete()
    assert le.line == "abc"
    assert le.value == "abc"

    le.press_delete()
    assert le.line == "abc"
    assert le.value == "abc"

    le.press_home()
    le.press_delete(5)
    assert le.line == ""
    assert le.value == ""


def _test_press_right():
    le = LineEditor()
    le.enter_text("alice")

    le.press_right()
    assert le.line == "alice"
    assert le.value == "alice"

    le.press_home()
    le.press_right()
    assert le.line == "alice"
    assert le.value == "a"


def _test_press_end():
    le = LineEditor()
    le.enter_text("alice")

    le.press_home()
    le.press_end()
    assert le.line == "alice"
    assert le.value == "alice"

    le.press_left()
    le.press_end()
    assert le.line == "alice"
    assert le.value == "alice"


def _test_clear_line():
    le = LineEditor()
    le.enter_text("alice")
    le.clear_line()
    assert le.line == ""
    assert le.value == ""

    le.enter_text("alice")
    le.press_left()
    le.clear_line()
    assert le.line == ""
    assert le.value == ""


def _test_clear_backward():
    le = LineEditor()
    le.enter_text("alice")
    le.press_left(2)
    le.clear_backward()
    assert le.line == "ce"
    assert le.value == ""


def _test_clear_forward():
    le = LineEditor()
    le.enter_text("alice")
    le.press_left(3)
    le.clear_forward()
    assert le.line == "al"
    assert le.value == "al"


def _test_move_word():
    le = LineEditor()
    text = " hello world alice bob "

    # --- move backward
    le.replace_text(text)
    le.move_word_backward()
    assert le.line == text
    assert le.value == " hello world alice "

    le.replace_text(text)
    le.press_left(1)
    le.move_word_backward()
    assert le.line == text
    assert le.value == " hello world alice "

    le.replace_text(text)
    le.press_left(2)
    le.move_word_backward()
    assert le.line == text
    assert le.value == " hello world alice "

    le.replace_text(text)
    le.press_left(7)
    le.move_word_backward()
    assert le.line == text
    assert le.value == " hello world "

    le.replace_text(text)
    le.move_to_start()
    le.move_word_backward()
    assert le.line == text
    assert le.value == ""

    le.replace_text("    ")
    le.move_word_backward()
    assert le.line == "    "
    assert le.value == ""

    # --- move forward
    le.replace_text(text)
    le.move_to_start()
    le.move_word_forward()
    assert le.line == text
    assert le.value == " hello"

    le.replace_text(text)
    le.move_to_start()
    le.press_right(1)
    le.move_word_forward()
    assert le.line == text
    assert le.value == " hello"

    le.replace_text(text)
    le.move_to_start()
    le.press_right(2)
    le.move_word_forward()
    assert le.line == text
    assert le.value == " hello"

    le.replace_text(text)
    le.move_to_start()
    le.press_right(7)
    le.move_word_forward()
    assert le.line == text
    assert le.value == " hello world"

    le.replace_text(text)
    le.move_to_end()
    le.move_word_forward()
    assert le.line == text
    assert le.value == text

    le.replace_text("    ")
    le.move_word_forward()
    assert le.line == "    "
    assert le.value == "    "


def test():
    print("")
    _test_is_cursor_at_begin_or_end_of_line()
    _test_press_key()
    _test_enter_text()
    _test_press_backspace()
    _test_press_left()
    _test_press_home()
    _test_press_delete()
    _test_press_right()
    _test_press_end()
    _test_clear_line()
    _test_clear_backward()
    _test_clear_forward()
    _test_move_word()


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "zelfred.line_editor", preview=False)



# -*- coding: utf-8 -*-

from zelfred.paths import dir_project_root
from zelfred.debug import Debugger


class TestDebugger:
    def test(self):
        debugger = Debugger()
        debugger.path_log_txt = dir_project_root / "tmp" / "log.txt"
        debugger.path_log_txt.parent.mkdir(parents=True, exist_ok=True)

        debugger.reset()
        debugger.enable()
        debugger.log("hello")
        assert debugger.path_log_txt.read_text().endswith("hello\n")

        debugger.disable()
        debugger.log("hello")
        assert debugger.path_log_txt.read_text().endswith("hello\n")

        debugger.enable()
        debugger.log("world")
        assert debugger.path_log_txt.read_text().endswith("world\n")


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "zelfred.debug", preview=False)



# -*- coding: utf-8 -*-

from zelfred import api


def test():
    _ = api
    _ = api.Item
    _ = api.T_ITEM
    _ = api.Query
    _ = api.QueryParser
    _ = api.UI
    _ = api.T_HANDLER
    _ = api.debugger
    _ = api.exc


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "afwf_shell.api", preview=False)



# -*- coding: utf-8 -*-

"""
You can use one of the following hotkey to do anything using the selected item.

- :meth:`Enter <zelfred.item.Item.enter_handler>`
- :meth:`Ctrl + A <zelfred.item.Item.ctrl_a_handler>`
- :meth:`Ctrl + W <zelfred.item.Item.ctrl_w_handler>`
- :meth:`Ctrl + P <zelfred.item.Item.ctrl_p_handler>`

This module provides some common actions that you can use directly.
"""

import subprocess
from pathlib import Path

try:
    from mac_notifications import client as mac_notification_client
except ImportError:
    pass

from .vendor.os_platform import IS_WINDOWS


def open_url(url: str):
    """
    Open a URL in the default browser.
    """
    if IS_WINDOWS:
        subprocess.run(["start", url], shell=True)
    else:
        subprocess.run(["open", url])


def open_file(path: Path):
    """
    Open a file in the default application.
    """
    if IS_WINDOWS:
        subprocess.run([str(path)], shell=True)
    else:
        subprocess.run(["open", str(path)])


def send_mac_notification(
    title: str,
    subtitle: str,
):
    """
    Send a MAC notification.

    This feature is based on the
    `macos-notifications <https://github.com/Jorricks/macos-notifications>`_
    Python library.

    However, this is not working on MacOS > 11.0, because of the API this library
     is using is deprecated. See this discussion for more details
     https://github.com/Jorricks/macos-notifications/issues/8.
    """
    mac_notification_client.create_notification(
        title=title,
        subtitle=subtitle,
    )



# -*- coding: utf-8 -*-

import typing as T
import readchar


class Event:
    pass


class KeyPressedEvent(Event):
    def __init__(self, value):
        self.value = value


class RepaintEvent(Event):
    pass


class KeyEventGenerator:
    def __init__(
        self,
        key_generator: T.Optional[T.Callable[[], str]] = None,
    ):
        self._key_generator = key_generator or readchar.readkey

    def next(self) -> KeyPressedEvent:
        return KeyPressedEvent(value=self._key_generator())



# -*- coding: utf-8 -*-

"""
Zelfred public API.

Usage example:

.. code-block:: python

    import zelfred.api as zf
"""

from .item import Item
from .item import T_ITEM
from .query import Query
from .query import QueryParser
from .ui import UI
from .ui import T_HANDLER
from .action import open_url
from .action import open_file
from .debug import debugger
from . import exc



# -*- coding: utf-8 -*-

from zelfred.query import Query, QueryParser


class TestQueryParser:
    def test_parse(self):
        qp = QueryParser(delimiter="/")
        q = qp.parse(" bucket / artifacts / / deploy.zip")
        assert q.parts == [" bucket ", " artifacts ", " ", " deploy.zip"]
        assert q.trimmed_parts == ["bucket", "artifacts", "deploy.zip"]

        qp = QueryParser(delimiter=[" ", "-", "_"])
        q = qp.parse(" a b-c d_e f-g_h ")
        assert q.trimmed_parts == list("abcdefgh")


class TestQuery:
    def test_from_str(self):
        q = Query.from_str("  a   b   c  ")
        assert q.trimmed_parts == list("abc")


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "zelfred.query", preview=False)



# -*- coding: utf-8 -*-

import pytest

from zelfred.dropdown import Dropdown
from zelfred.exc import NoItemToSelectError


def get_items(menu) -> str:
    return "".join([item for item, flag in menu])


def get_selected_item(menu) -> str:
    for item, flag in menu:
        if flag:
            return item


def _test_high_amount_of_items():
    dd = Dropdown(items=list("abcdefghijklmnopqrstuvwxyz"))
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    dd.press_up()
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    dd.scroll_up(2)
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    menu = dd.menu
    assert get_items(menu) == "abcdefghij"
    assert get_selected_item(menu) == "a"

    dd.press_down(5)
    assert dd.selected_item_index == 5
    assert dd.cursor_position == 5

    menu = dd.menu
    assert get_items(menu) == "abcdefghij"
    assert get_selected_item(menu) == "f"

    dd.press_up(2)
    assert dd.selected_item_index == 3
    assert dd.cursor_position == 3

    menu = dd.menu
    assert get_items(menu) == "abcdefghij"
    assert get_selected_item(menu) == "d"

    dd.scroll_down(2)
    assert dd.selected_item_index == 13
    assert dd.cursor_position == 9

    menu = dd.menu
    assert get_items(menu) == "efghijklmn"
    assert get_selected_item(menu) == "n"

    dd.press_up(2)
    assert dd.selected_item_index == 11
    assert dd.cursor_position == 7

    menu = dd.menu
    assert get_items(menu) == "efghijklmn"
    assert get_selected_item(menu) == "l"

    dd.scroll_down(2)
    assert dd.selected_item_index == 21
    assert dd.cursor_position == 9

    menu = dd.menu
    assert get_items(menu) == "mnopqrstuv"
    assert get_selected_item(menu) == "v"

    dd.scroll_down(2)
    assert dd.selected_item_index == 25
    assert dd.cursor_position == 9

    menu = dd.menu
    assert get_items(menu) == "qrstuvwxyz"
    assert get_selected_item(menu) == "z"

    dd.press_down(2)
    assert dd.selected_item_index == 25
    assert dd.cursor_position == 9

    menu = dd.menu
    assert get_items(menu) == "qrstuvwxyz"
    assert get_selected_item(menu) == "z"

    dd.press_up(3)
    assert dd.selected_item_index == 22
    assert dd.cursor_position == 6

    menu = dd.menu
    assert get_items(menu) == "qrstuvwxyz"
    assert get_selected_item(menu) == "w"


def _test_low_amount_of_items():
    dd = Dropdown(items=list("abc"))
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    dd.press_up()
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    dd.scroll_up(2)
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "a"

    dd.press_down(2)
    assert dd.selected_item_index == 2
    assert dd.cursor_position == 2

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "c"

    dd.press_up(1)
    assert dd.selected_item_index == 1
    assert dd.cursor_position == 1

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "b"

    dd.scroll_down(2)
    assert dd.selected_item_index == 2
    assert dd.cursor_position == 2

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "c"

    dd.press_up(2)
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "a"

    dd.scroll_down(2)
    assert dd.selected_item_index == 2
    assert dd.cursor_position == 2

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "c"

    dd.press_down(2)
    assert dd.selected_item_index == 2
    assert dd.cursor_position == 2

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "c"

    dd.press_up(1)
    assert dd.selected_item_index == 1
    assert dd.cursor_position == 1

    menu = dd.menu
    assert get_items(menu) == "abc"
    assert get_selected_item(menu) == "b"


def _test_press_up_and_down():
    dd = Dropdown(items=list("abc"))
    dd._press_up()
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0
    assert dd.selected_item == "a"

    dd.press_down(3)
    dd._press_down()
    assert dd.selected_item_index == 2
    assert dd.cursor_position == 2
    assert dd.selected_item == "c"


def _test_update():
    dd = Dropdown(items=list("abc"))
    dd._press_down()
    dd.update(items=list("mn"))
    assert dd.items == list("mn")
    assert dd.n_items == 2
    assert dd.selected_item_index == 0
    assert dd.cursor_position == 0
    assert dd.show_items_limit == 2


def _test_selected_item():
    dd = Dropdown(items=[])
    with pytest.raises(NoItemToSelectError):
        _ = dd.selected_item


def test():
    _test_high_amount_of_items()
    _test_low_amount_of_items()
    _test_press_up_and_down()
    _test_update()
    _test_selected_item()


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "zelfred.dropdown", preview=False)



# -*- coding: utf-8 -*-

from zelfred.item import T_ITEM, Item, DEFAULT_TITLE, DEFAULT_SUBTITLE


class TestItem:
    def test(self):
        # test title and subtitle
        item = Item(title="hello")
        assert item.title_text == "hello"
        assert item.subtitle_text == DEFAULT_SUBTITLE

        item = Item(title="", subtitle="world")
        assert item.title_text == DEFAULT_TITLE
        assert item.subtitle_text == "world"

        # default handler should have no effect.
        item.enter_handler(ui=None)
        item.ctrl_a_handler(ui=None)
        item.ctrl_w_handler(ui=None)
        item.ctrl_p_handler(ui=None)

        # default uid should be an uuid
        item = Item(title="hello")
        assert len(item.uid) == 32

        item1 = Item(title="hello")
        item2 = Item(title="hello")
        assert item1.uid != item2.uid


if __name__ == "__main__":
    from zelfred.tests import run_cov_test

    run_cov_test(__file__, "zelfred.item", preview=False)



# -*- coding: utf-8 -*-

import typing as T


class EndOfInputError(Exception):
    """ """

    def __init__(
        self,
        selection: T.Any,
        *args,
    ):
        super().__init__(*args)
        self.selection = selection


class JumpOutLoopError(Exception):
    pass


class TerminalTooSmallError(SystemError):
    pass


class NoItemToSelectError(IndexError):
    pass





# -*- coding: utf-8 -*-

"""
Item is the basic building block of the terminal UI. It represents a single
selectable item in the dropdown menu. User can move or scroll selector up and down
to select an item, and then perform :mod:`~zelfred.action`.
"""

import typing as T
import uuid
import dataclasses

from .constants import DEFAULT_TITLE, DEFAULT_SUBTITLE

if T.TYPE_CHECKING:  # pragma: no cover
    from .ui import UI


def gen_uid() -> str:
    return uuid.uuid4().hex


@dataclasses.dataclass
class Item:
    """
    Represent an item in the terminal UI. In the example below, the
    ``[x] Beautiful is better than ugly`` and ``subtitle 01`` is one item.

    .. code-block:: bash

        (Query):
        [x] Beautiful is better than ugly.
              subtitle 01
        [ ] Explicit is better than implicit.
              subtitle 02
        [ ] Simple is better than complex.
              subtitle 03

    :param uid: item unique id. The UI use this to distinguish different items.
    :param title: first line of the item. It has a checkbox in front of it to
        indicate whether it is selected.
    :param subtitle: second line of the item.
    :param arg: argument that will be passed to the action.
    :param autocomplete: the text that will be filled in the input box when
        user hits ``TAB`` key.
    :param variables: arbitrary dictionary object, it can be used in the action.
    """

    title: str = dataclasses.field()
    subtitle: T.Optional[str] = dataclasses.field(default=None)
    uid: str = dataclasses.field(default_factory=gen_uid)
    arg: T.Optional[str] = dataclasses.field(default=None)
    autocomplete: T.Optional[str] = dataclasses.field(default=None)
    variables: T.Dict[str, T.Any] = dataclasses.field(default_factory=dict)

    @property
    def title_text(self) -> str:
        return self.title or DEFAULT_TITLE

    @property
    def subtitle_text(self) -> str:
        return self.subtitle or DEFAULT_SUBTITLE

    def enter_handler(self, ui: "UI"):
        """
        This is the abstract method that will perform user defined action
        when user hits ``Enter`` on this item. Develop should inherit this class
        and override this method to perform user defined action.

        :param ui: the :class:`~afwf_shell.ui.UI` object.
        """
        pass

    def ctrl_a_handler(self, ui: "UI"):
        """
        This is the abstract method that will perform user defined action
        when user hits ``Ctrl + A`` on this item. Develop should inherit this class
        and override this method to perform user defined action.

        :param ui: the :class:`~afwf_shell.ui.UI` object.
        """
        pass

    def ctrl_w_handler(self, ui: "UI"):
        """
        This is the abstract method that will perform user defined action
        when user hits ``Ctrl + W`` on this item. Develop should inherit this class
        and override this method to perform user defined action.

        :param ui: the :class:`~afwf_shell.ui.UI` object.
        """
        pass

    def ctrl_p_handler(self, ui: "UI"):
        """
        This is the abstract method that will perform user defined action
        when user hits ``Ctrl + P`` on this item. Develop should inherit this class
        and override this method to perform user defined action.

        :param ui: the :class:`~afwf_shell.ui.UI` object.
        """
        pass


T_ITEM = T.TypeVar("T_ITEM", bound=Item)



# -*- coding: utf-8 -*-

# number of item can be shown in the UI
SHOW_ITEMS_LIMIT = 10

# how many UP/DOWN key pressed when scrolling the UI.
SCROLL_SPEED = 5

DEFAULT_TITLE = "NO TITLE"
DEFAULT_SUBTITLE = "NO SUBTITLE"
